!!****if* source/Simulation/SimulationMain/Sod/Simulation_initBlock
!!
!! NAME
!!
!!  Simulation_initBlock
!!
!!
!! SYNOPSIS
!!
!!  Simulation_initBlock(integer(IN) :: blockID)
!!
!!
!!
!! DESCRIPTION
!!
!!  Initializes fluid data (density, pressure, velocity, etc.) for
!!  a specified block.  This version uses particle sph data to set up
!!  the block.
!!
!!
!! ARGUMENTS
!!
!!  blockID -           the number of the block to update
!!
!! PARAMETERS
!!    ???
!!
!! GLOBAL VARIABLES
!!  sim_npart
!!  sim_partPos(npart,3)
!!  sim_partVel(npart,3)
!!  sim_partH(npart)
!!  sim_partM(npart)
!!  sim_partU(npart)
!!
!!  sim_xar36
!!  sim_xc12
!!  sim_xca40
!!  sim_xcr48
!!  sim_xfe52
!!  sim_xhe4
!!  sim_xmg24
!!  sim_xne20
!!  sim_xni56
!!  sim_xo16
!!  sim_xs32
!!  sim_xsi28
!!  sim_xti44
!!
!! LOCAL VARIABLES
!!  pos(3)
!!  vel(3)
!!  h
!!  m
!!  f
!!  var(?)
!!
!!  xCen(sizeX)
!!  yCen(sizeY)
!!  zCen(sizeZ)
!!
!!  width(3)
!!
!!  inside
!!
!!  blkLow(3), blkHigh(3)
!! 
!!  rhoZone, velZone(3), presZone, eintZone, enerZone, ekinZone
!! 
!!***

subroutine Simulation_initBlock(blockID)

  use Simulation_data
  use Grid_data, ONLY: gr_smalle

  use Driver_interface, ONLY: Driver_abortFlash
 
  use Grid_interface, ONLY : Grid_getBlkIndexLimits, Grid_getCellCoords, &
                             Grid_getPointData, Grid_putPointData, &
                             Grid_getBlkPtr, Grid_releaseBlkPtr, &
                             Grid_getBlkRefineLevel

  use Eos_interface, ONLY : Eos
  use Eos_data, ONLY : eos_gamma

  use Logfile_interface, ONLY : Logfile_stamp
  use tree, ONLY : lrefine_max

  implicit none

#include "constants.h"
#include "Flash.h"
#include "Eos.h"
  
  integer, intent(in) :: blockID

  real, parameter :: Boltzmann = 1.3807e-16
! Dummy indices / values
  integer :: i, j, k, n
  real :: q1, q2, q3, q4, qi,qf

! Particular particle values
  real, dimension(MDIM) :: pos, vel

! Grid values
  real :: width
  real,allocatable, dimension(:) :: xCen, yCen, zCen
  integer :: sizeX,sizeY,sizeZ, deltaCell
  integer, dimension(2,MDIM) :: blkLimits, blkLimitsGC
  real, dimension(MDIM) :: blkLow, blkHigh

! Values to be put in cell  
  real :: rhoZone,  presZone, eintZone, enerZone, ekinZone, tempZone, gammZone, spdsqrd, tempRadZone, tempEleZone, &
          presRadZone, presEleZone, eintRadZone, eintEleZone
  real, dimension(MDIM) :: velZone
  integer, dimension(MDIM) :: posCell
  real, dimension(:,:,:,:),pointer :: solnData
  
! Variables for Eos
  real, dimension(EOS_NUM) :: eosData
  integer eosMode

  logical :: gcell = .true., inside, inrange
  integer :: istat 

  real :: smallNum = 1e-10
  integer :: refine_level

  ! get the integer index information for the current block
  call Grid_getBlkIndexLimits(blockId,blkLimits,blkLimitsGC)
  call Grid_getBlkRefineLevel(blockID,refine_level) 
 
  ! Determine size of interior of block
  sizeX = blkLimitsGC(HIGH,IAXIS) - blkLimitsGC(LOW,IAXIS) + 1
  sizeY = blkLimitsGC(HIGH,JAXIS) - blkLimitsGC(LOW,JAXIS) + 1
  sizeZ = blkLimitsGC(HIGH,KAXIS) - blkLimitsGC(LOW,KAXIS) + 1

  ! allocate memory for iCen variables
  allocate(xCen(sizeX),stat=istat)
  allocate(yCen(sizeY),stat=istat)
  allocate(zCen(sizeZ),stat=istat)

  ! Initialize
  xCen = 0.0d0
  yCen = 0.0d0
  zCen = 0.0d0

  ! Get interior cell coordinates
  if (NDIM == 3) &
    call Grid_getCellCoords(KAXIS, blockId, CENTER ,gcell, zCen, sizeZ)

  if (NDIM >= 2) &
    call Grid_getCellCoords(JAXIS, blockId, CENTER,gcell, yCen, sizeY)

  call Grid_getCellCoords(IAXIS, blockId, CENTER,gcell, xCen, sizeX)
 
  ! Get lower and upper coordinates
  blkLow(1)   = xCen(1)
  blkHigh(1) = xCen(sizeX)

  blkLow(2)   = yCen(1)
  blkHigh(2) = yCen(sizeY)

  blkLow(3)   = zCen(1)
  blkHigh(3) = zCen(sizeZ)
 
  ! Get cell dimensions; currently required to be cubic
  width = xCen(2) - xCen(1)

  q4 = abs(width-(yCen(2)-yCen(1)))
  if((q4.gt.smallNum*width).and.(NDIM>1)) then
    write(6,*) "Non-symmetric cell dimensions"
    write(6,*) "Width = ", width, " while new y width = ",yCen(2)-yCen(1)
    call Driver_abortFlash('Non symmetric cell dimensions. See output file for more details')
  endif
 
  q4 = abs(width-(zCen(2)-zCen(1)))
  if((q4.gt.smallNum*width).and.(NDIM>2)) then
    write(6,*) "Non-symmetric cell dimension"
    write(6,*) "Width = ", width, " while new z width = ",zCen(2)-zCen(1) 
    call Driver_abortFlash('Non symmetric cell dimensions. See output file for more details')
  endif

  ! Get pointer to unk
  call Grid_getBlkPtr(blockId, solnData)


  ! Take note of some confusion:
  !   - Cell positions are ONLY found for the interior (not including GC)
  !   - solnData points to all cells INCLUDING GC
  !   - posCell will have to use the GC grid if it will later be used with solnData
  !      - thus for comparisons between posCell and iCen, iCen must have the same
  !        index as posCell minus blkLimits(DIRECTION) + 1

!------------------------------------------------------------------------------
! Loop over all particle positions. If particle's smoothing length is within the
! the block, then it's attributes are applied to its neighbouring cells, subject to
! the particle's smoothing kernel.

  solnData = 0.d0
  if (sim_gradDir .eq. 1) then
    qi = xCen(1)
    qf = xCen(sizeX)
  else if (sim_gradDir .eq. 2) then
    qi = yCen(1)
    qf = yCen(sizeY)
  else
    qi = zCen(1)
    qf = zCen(sizeZ)
  endif
  do k = blkLimitsGC(LOW,KAXIS),blkLimitsGC(HIGH,KAXIS)
     q3 = zCen(k-blkLimitsGC(LOW,KAXIS)+1)
     
     do j = blkLimitsGC(LOW,JAXIS),blkLimitsGC(HIGH,JAXIS)
        q2 = yCen(j-blkLimitsGC(LOW,JAXIS)+1)
        
        do i = blkLimitsGC(LOW,IAXIS),blkLimitsGC(HIGH,IAXIS)
          q1 = xCen(i-blkLimitsGC(LOW,IAXIS)+1)

          if (sim_gradDir .eq. 1) then
            rhoZone = ( (q1-qi)*sim_rhoAmb2 + (qf-q1)*sim_rhoAmb)/(qf-qi)
          else if (sim_gradDir .eq. 2) then
            rhoZone = ( (q2-qi)*sim_rhoAmb2 + (qf-q2)*sim_rhoAmb)/(qf-qi)
          else 
            rhoZone = ( (q3-qi)*sim_rhoAmb2 + (qf-q3)*sim_rhoAmb)/(qf-qi)
          endif
          if (sim_velDir .eq. 1) then
            solnData(VELX_VAR,i,j,k) = sim_vdrift    
          else if (sim_velDir .eq. 2) then
            solnData(VELY_VAR,i,j,k) = sim_vdrift    
          else 
            solnData(VELZ_VAR,i,j,k) = sim_vdrift    
          endif
          eintZone = sim_eintAmb
          presZone = sim_presAmb
          tempZone = sim_tempAmb
          ekinZone = 0.5d0*sim_vdrift**2
          solnData(DENS_VAR,i,j,k) = rhoZone 

          eosMode = MODE_DENS_PRES          
          enerZone = ekinZone + eintZone

          eosData(EOS_DENS) = rhoZone
          eosData(EOS_PRES) = presZone
                   
          call Eos(eosMode, 1, eosData)
          if ( rhoZone /=  eosData(EOS_DENS) ) stop "Density was updated - things need to be renormalised"

          eintZone = eosData(EOS_EINT)
          presZone = eosData(EOS_PRES)
          tempZone = eosData(EOS_TEMP)
          gammZone = eosData(EOS_GAMC)
          enerZone = eintZone + ekinZone

          if (presZone .ne. sim_presAmb) then
            write(6,*) "Error: presZone not equal to presAMB: ", presZone, sim_presAmb
            call Driver_abortFlash("Sim_initBlock ERROR: presZone isn't equal to presAMB.")
          endif

          solnData(EINT_VAR,i,j,k) = eintZone
          solnData(PRES_VAR,i,j,k) = presZone
          solnData(TEMP_VAR,i,j,k) = tempZone
          solnData(ENER_VAR,i,j,k) = enerZone
          solnData(GAMC_VAR,i,j,k) = gammZone
          solnData(GAME_VAR,i,j,k) = gammZone

        enddo
     enddo
  enddo


!! Cleanup!  Must deallocate arrays
  deallocate(xCen)
  deallocate(yCen)
  deallocate(zCen)

  call Grid_releaseBlkPtr(blockID, solnData)
 
  return
end subroutine Simulation_initBlock




!
!
!

subroutine findFact(xP,yP,zP,xo,yo,zo,hP,f,L, d,order)

! ==================================================================== c
!   x,y,z = particle's position
!   xo,yo,zo = position of one cell center (usually the inner(i,j,k) cell
!   h     = particle's smoothing length
!   f     = particle's mass conservation factor
!   L     = uniform grid spacing
!   d     = number of cells in radius reached by 2h
!   order = 1 --> first
!         = 2 --> second
! ==================================================================== c
!
!   DECLARATIONS
! 
! ==================================================================== c

  implicit none

  real :: xP,yP,zP,x,y,z,xo,yo,zo,hP,h,f,L

  real :: w,r,mTot, q1, q2, q3

  integer i, j, k, n, d, order

! ==================================================================== c
!
!   INITIALIZATIONS
!
! ==================================================================== c
  mTot = 0.D0
! ==================================================================== c
!
!   Renormalization:
!                    Make the grid have unit spacing
! ==================================================================== c

  x = (xP-xo)/L
  y = (yP-yo)/L
  z = (zP-zo)/L
  h = hP/L

! ==================================================================== c
!
!   Centering:
!              Translate particle so it is in the cell centered
!              at (0,0,0)
! ==================================================================== c

  x = mod(x,1.D0)
  if (x.gt.0.5D0) x = x - 1.D0

  y = mod(y,1.D0)
  if (y.gt.0.5D0) y = y - 1.D0

  z = mod(z,1.D0)
  if (z.gt.0.5D0) z = z - 1.D0

! ==================================================================== c
!
!   Determine bounds: 
!                     We will only look at cells within 2h of the 
!                     central cell's edge.
! ==================================================================== c

  d = int(2.D0*h + 0.5D0)

! ==================================================================== c
!
!   Loop over cells and if distance between cell and particle is less
!   than 2h, then determine cell's density and add to total mass.
!
!   Note: since h was rescaled, w is a factor of L**3 higher but when
!         multiplying by the volume of the cell, this is canceled out.      
!
! ==================================================================== c

  do k=-d, d
    q3=(dble(k)-z)**2

    do j=-d,d
      q2=q3+(dble(j)-y)**2

      do i=-d,d
        q1=q2+(dble(i)-x)**2

        r = sqrt(q1)

        if(r .lt. 2.D0*h) then
          if (order .eq. 1) then
            mTot = mTot + w(r,h)
          else
            mTot = mTot + w(r,h)**2
          endif
        endif
      enddo

     enddo
 
  enddo

! ==================================================================== c
!
!   Find the factor f = m / mtot
!
! ==================================================================== c
  if (mTot.ne.0.D0) then
    f = 1.D0/mTot
  else
    f = 0.d0
  endif
! ==================================================================== c

  return
end subroutine findFact
!
!
!
! ==================================================================== c
! ==================================================================== c
! ==================================================================== c


real function w(r,h)

  implicit none
  real :: r, h, v

  v = r/h

  w = 3.1415927410125732d0*h*h*h
  w = 1.0d0/w

  if (v.lt.1.d0) then
     w = w*(1-1.5d0*v*v+0.75d0*v*v*v)
     return
  else if (v.lt.2.0d0) then
     w = w*0.25d0*(2.0d0-v)**3
     return
  else
     w = 0.D0
  endif
 
  return
end function w

! ==================================================================== c










